from pds4_general_objects import Structure, Meta_Class, Meta_Structure
from pds4_constants import PDS4_NAMESPACES
from pds4_helpers import find_xml_element
from pds4_exceptions import PDS4StandardsException

# Safe import of OrderedDict
try:
    from collections import OrderedDict
except ImportError:
    from ordered_dict_26 import OrderedDict


class ArrayStructure(Structure):
    """ ArrayStructure class; stores a single PDS4 array data structure

     PDS4 array structures are Array, Array_2D and derivatives, Array_3D and derivatives. This object
     includes the array's data, its meta data and its description in the label"""

    def info(self, abbreviated=False):

        abbreviated_info = "{0} '{1}' ({2} axes)".format(
                            self.type, self.id, len(self.meta_data.get_axis_arrays()))

        if abbreviated:
            print abbreviated_info

        else:
            print 'Axes for {0}: \n'.format(abbreviated_info)

            for axis in self.meta_data.get_axis_arrays():
                print '{0} ({1} elements)'.format(axis['axis_name'], axis['elements'])

    def is_displayable(self):

        has_display_dict = self.meta_data.display_settings is not None

        if ('Array_2D_' in self.type) or ('Array_3D_' in self.type) or has_display_dict:
            return True

        return False


class Meta_ArrayStructure(Meta_Structure):
    """ Stores meta data about a single Array structure (e.g. Array, Array_2D, Array_3D_Spectrum, etc) """

    def __init__(self, *args, **kwds):
        super(Meta_ArrayStructure, self).__init__(*args, **kwds)

        # Contains the Meta_DisplaySettings for this Array structure, if one exists
        self.display_settings = None

    def load(self, xml_array, full_label):

        self._load_keys_from_xml(xml_array)

        # Ensure required keys for Array_* exist
        keys_must_exist = ['offset', 'axes', 'Axis_Array', 'Element_Array']
        self._check_keys_exist(keys_must_exist)

        # Ensure required keys for Axis_Array(s) exist
        axis_keys_must_exist = ['axis_name', 'elements', 'sequence_number']
        multiple_axes = True if len(self.get_axis_arrays()) > 1 else False
        self._check_keys_exist(axis_keys_must_exist, sub_element='Axis_Array', is_sequence=multiple_axes)

        # Ensure required keys for Element_Array exist
        self._check_keys_exist(['data_type'], sub_element='Element_Array')

        # Add the Meta_DisplaySettings
        if 'local_identifier' in self:

            try:
                display_settings = Meta_DisplaySettings()
                display_settings.load(full_label, self['local_identifier'])
                self.display_settings = display_settings

            # The display dictionary is not guaranteed to exist
            except KeyError:
                pass

    # Convenience method to always return self['Axis_Array'] as a list of Arrays regardless of whether there
    # is a single Axis_Array or multiple
    def get_axis_arrays(self):

        axis_arrays = self['Axis_Array']

        if isinstance(axis_arrays, (list, tuple)):
            return list(axis_arrays)

        else:
            return [axis_arrays]

    # Convenience method to retrieve a self['Axis_Array'] by axis_name, sequence_number or both
    def get_axis_array(self, axis_name=None, sequence_number=None):

        axis_arrays = self.get_axis_arrays()
        retrieved_axis = None

        for axis in axis_arrays:

            # Find by both axis_name and sequence_number
            if (axis_name is not None) and (sequence_number is not None):

                if (axis['axis_name'] == axis_name) and (axis['sequence_number'] == sequence_number):
                    retrieved_axis = axis

            # Find by axis_name
            elif (axis_name is not None) and (axis['axis_name'] == axis_name):
                retrieved_axis = axis

            # Find by sequence_number
            elif (sequence_number is not None) and (axis['sequence_number'] == sequence_number):
                retrieved_axis = axis

        return retrieved_axis


class Meta_DisplaySettings(Meta_Class):
    """ Stores meta data about a Display Dictionary for a single Array structure """

    def __init__(self, *args, **kwds):
        super(Meta_DisplaySettings, self).__init__(*args, **kwds)

        # Set to True if Display Dictionary being used on load() is valid/supported
        self.valid = True

    def load(self, label, structure_lid):

        # Find the Display Dictionary matching structure_lid from the label
        matching_display = None

        displays = label.findall('.//disp:Display_Settings')
        if not displays:
            raise KeyError('Valid/supported Display Settings were not found')

        for display in displays:

            lid = find_xml_element(display, './/disp:local_identifier_reference')
            if lid == structure_lid:
                matching_display = display

        if matching_display is None:
            raise KeyError("Display Settings with for the LID '{0}' was not found".format(structure_lid))

        # Find structure being referenced by structure_lid
        found_structures = label.findall('.//File_Area_Observational/*')
        structure = None

        for found_structure in found_structures:

            found_lid = find_xml_element(found_structure, 'local_identifier')

            if (found_lid is not None) and (found_lid == structure_lid):
                structure = found_structure

        if structure is None:
            raise PDS4StandardsException("A Data Structure having the LID '{0}', specified in "
                                         "Display Settings, was not found".format(structure_lid))

        tag_modify = ('{{{0}}}'.format(PDS4_NAMESPACES['disp']), '')
        self._load_keys_from_xml(matching_display, tag_modify=tag_modify)

        try:
            self.validate(structure)
            self.valid = True

        # The display dictionary is not guaranteed to valid (unsupported older versions exist)
        except KeyError:
            self.valid = False

    def validate(self, xml_array):

        structure_lid = xml_array.find('local_identifier')

        # Ensure required keys for Display_Settings exist
        keys_must_exist = ['Local_Internal_Reference', 'Display_Direction']
        self._check_keys_exist(keys_must_exist)

        # Ensure required keys for Local_Internal_Reference exist
        reference_keys_must_exist = ['local_identifier_reference', 'local_reference_type']
        self._check_keys_exist(reference_keys_must_exist, sub_element='Local_Internal_Reference')

        # Ensure required keys for Display_Direction exist
        display_keys_must_exist = ['horizontal_display_axis', 'horizontal_display_direction',
                                   'vertical_display_axis', 'vertical_display_direction']
        self._check_keys_exist(display_keys_must_exist, sub_element='Display_Direction')

        # Ensure required keys for Color_Display_Settings exists, if the class exists
        if 'Color_Display_Settings' in self:

            color_keys_must_exist = ['color_display_axis', 'red_channel_band', 'green_channel_band',
                                     'blue_channel_band']
            self._check_keys_exist(color_keys_must_exist, sub_element='Color_Display_Settings')

        # Ensure required key for Movie_Display_Settings exists, if the class exists
        if 'Movie_Display_Settings' in self:
            self._check_keys_exist(['time_display_axis'], sub_element='Movie_Display_Settings')

        # Ensure required axes referenced by the Display Dictionary actually exist in the structure
        axes_arrays = xml_array.findall('Axis_Array')
        display_direction = self['Display_Direction']
        color_settings = self.get('Color_Display_Settings')
        movie_settings = self.get('Movie_Display_Settings')

        horizontal_axis_exists = False
        vertical_axis_exists = False
        color_axis_exists = False
        movie_axis_exists = False

        for axis in axes_arrays:

            axis_name = find_xml_element(axis, 'axis_name')

            if axis_name == display_direction['horizontal_display_axis']:
                horizontal_axis_exists = True

            if axis_name == display_direction['vertical_display_axis']:
                vertical_axis_exists = True

            if (color_settings is not None) and (axis_name == color_settings['color_display_axis']):
                color_axis_exists = True

            if (movie_settings is not None) and (axis_name == movie_settings['time_display_axis']):
                movie_axis_exists = True

        display_axes_error = (not horizontal_axis_exists) or (not vertical_axis_exists)
        color_axis_error = (color_settings is not None) and (not color_axis_exists)
        movie_axis_error = (movie_settings is not None) and (not movie_axis_exists)

        if display_axes_error or color_axis_error or movie_axis_error:
            raise PDS4StandardsException("An axis_name, specified in the Display Dictionary for LID '{0}', "
                                         "was not found".format(structure_lid))