import array

# Safe import of numpy (not required)
try:
    import numpy as np
except ImportError:
    np = None


# Creates a DataList, DataArray or DataNumpy based on input
def get_data_class(data, meta_data, use_numpy=False):

    if use_numpy:
        return DataNumpy(data, meta_data)

    elif isinstance(data, list):
        return DataList(data, meta_data)

    elif isinstance(data, array.array):
        return DataArray(data, meta_data)

    else:

        raise TypeError('Unknown type of Data found: {0}'.format(type(data)))


class DataList(list):
    """ DataList class; subclasses `list` and adds a meta_data attribute

    Stores PDS4 string and complex data. May also store numeric data exceeding sizes available with
    array.array when use_numpy is False, or sizes exceeding numpy dtypes when use_numpy is True. """

    def __init__(self, data, meta_data):
        super(DataList, self).__init__(data)

        self.meta_data = meta_data

    # Returns a pure List, without copying. It should very rare to ever need this because DataList allows
    # for all things that List allows
    def to_list(self):
        return self[:]


class DataArray(array.array):
    """ DataArray class; subclasses `array.array` and adds a meta_data attribute

    Stores PDS4 numeric data, except for complex data and numeric data exceeding sizes available with
    `array.array` when use_numpy is False. """

    def __new__(cls, data, meta_data):

        obj = array.array.__new__(cls, data.typecode, data)
        obj.meta_data = meta_data

        return obj

    # Returns a pure array.array, without copying. It should very rare to ever need this because DataArray
    # allows for all things that array allows
    def to_array(self):
        return self[:]

# Try is used for this class declaration because numpy is not guaranteed to be available
try:

    class DataNumpy(np.ndarray):
        """ DataNumpy class; subclasses `ndarray` and adds a meta_data attribute

        Stores PDS4 data when use_numpy is True, except for numeric data exceeding sizes available with
        numpy dtypes. """

        def __new__(cls, data, meta_data):

            obj = np.asanyarray(data).view(cls)
            obj.meta_data = meta_data

            return obj

        def __array_finalize__(self, obj):

            if obj is None:
                return

            self.meta_data = getattr(obj, 'meta_data', None)

        # Returns a pure np.ndarray, without copying. It should very rare to ever need this because DataNumpy
        # allows for all things that np.ndarray allows
        def to_ndarray(self):
            return self.view(np.ndarray)

# Define a dummy, unused, DataNumpy so it can be imported without error if numpy is unavailable
except AttributeError:

    class DataNumpy(object):

        def __init__(self, data, meta_data):
            pass

        pass

    pass