import re
from collections import Sequence
from itertools import chain
import xml.etree.ElementTree as ET

from pds4_read_label import read_label
from pds4_helpers import xml_to_dict
from pds4_constants import PDS4_NAMESPACES
from pds4_exceptions import PDS4StandardsException
from pds4_logging import logger_init

# Safe import of OrderedDict
try:
    from collections import OrderedDict
except ImportError:
    from ordered_dict_26 import OrderedDict

# Initialize the logger
logger = logger_init()

#################################


class StructureList(Sequence):
    """ StructureList class; stores label and all data structures of a PDS4 observational product

    Allows both Dict-like and List-like access. This class is returned by read_pds4().
    """

    def __init__(self, structures, label, read_in_log):
        super(StructureList, self).__init__()

        self.structures = structures

        self.label = label
        self.read_in_log = read_in_log

    def __getitem__(self, key):

        if isinstance(key, (int, long)) or isinstance(key, slice):
            structure = self.structures[key]

        else:

            # Reuse tuple-search logic for simple (ie, str) name and lid searches
            if not isinstance(key, tuple):
                key = (key,) + (0,)

            # Search by local identifier
            structure = self._get_structure_by_id(key, 'local_identifier')

            # Search by name
            if structure is None:
                structure = self._get_structure_by_id(key, 'name')

        return structure

    def __len__(self):
        return len(self.structures)

    def info(self, abbreviated=True):

        for structure in self.structures:

            if abbreviated:
                structure.info(abbreviated=True)

            else:
                structure.info(abbreviated=False)
                print '---------------------------------------- \n'

    def _get_structure_by_id(self, key, id_type):

        structure = None

        structure_lids = [obj.meta_data.get(id_type) for obj in self.structures]
        structure_match_idx = [i for i, lid in enumerate(structure_lids) if lid == key[0]]

        if len(structure_match_idx) > key[1]:
            structure = self.structures[structure_match_idx[key[1]]]

        return structure


class Structure(object):
    """ Base class; stores a single PDS4 data structure (e.g. Array, or Table)

     Includes the data structure's data, its meta data and its description in the label"""

    def __init__(self, parent_filename, structure_label, structure_meta_data, structure_data=None):

        super(Structure, self).__init__()

        self.id = structure_meta_data.id
        self.type = structure_label.tag

        self.parent_filename = parent_filename
        self.label = structure_label
        self.meta_data = structure_meta_data

        self.data = structure_data
        if structure_data is None:
            self.data = []

    def info(self, abbreviated=False):
        print NotImplementedError

    def is_array(self):

        if 'Array' in self.type:
            return True

    def is_table(self):

        if 'Table' in self.type:
            return True


class Meta_Class(OrderedDict):
    """ Base class; subclassed by any Meta_* class """

    # Checks if the keys exist in self. If sub_element is set, then self[sub_element] is checked for keys
    # instead. If is_sequence, then the self or self[sub_element] is a sequence, and each element of the
    # sequence is checked for keys
    def _check_keys_exist(self, keys, sub_element=None, is_sequence=False):

        # Determine where to look for keys
        if sub_element is not None:
            struct = self[sub_element]
            error_addition = "'{0}' in ".format(sub_element)
        else:
            struct = self
            error_addition = ''

        # Check for the existence of each key
        for key in keys:

            error = "{0}{1} must have an '{2}'".format(error_addition, type(self).__name__, key)

            if is_sequence:

                # Recurse into struct if it's a sequence
                for element in struct:

                    if key not in element:
                        raise PDS4StandardsException(error)
            else:

                if key not in struct:
                    raise PDS4StandardsException(error)

    # Load keys from XML
    def _load_keys_from_xml(self, xml, cast_values=True, tag_modify=()):

        items = xml_to_dict(xml, cast_values=cast_values, skip_attributes=True,
                            tag_modify=tag_modify).values()[0]

        for key, value in items.iteritems():
            self[key] = value


class Meta_Structure(Meta_Class):
    """ Base class; subclassed by Meta_TableStructure and Meta_ArrayStructure

    Contains the meta data for a PDS4 data structure"""

    # Get the ID, either from name or local_identifier, if either one exists
    @property
    def id(self):

        id = self.get('local_identifier')

        if id is None:
            id = self.get('name')

        return id


class Label(object):
    """ Label class; contains a PDS4 label """

    # Init. Methods in this class expect convenient_root and unmodified_root to have been read-in
    # via pds4_read_label(), not via ET directly
    def __init__(self, convenient_root=None, unmodified_root=None):

        # ElementTree root containing read-in XML which was modified before storage (see options set in load() below)
        self._convenient_root = convenient_root

        # ElementTree root containing the unmodified XML
        self._unmodified_root = unmodified_root

    def __getitem__(self, key):
        return Label(self._convenient_root[key], self._unmodified_root[key])

    def load(self, filename):

        self._convenient_root = read_label(filename, strip_extra_whitespace=True, enforce_default_prefixes=True)

        self._unmodified_root = read_label(filename, strip_extra_whitespace=False, enforce_default_prefixes=False)

    @property
    def text(self):
        return self.getroot().text

    @property
    def tail(self):
        return self.getroot().tail

    @property
    def tag(self):
        return self.getroot().tag

    @property
    def attrib(self):
        return self.getroot().attrib

    # Returns root element of label
    def getroot(self, unmodified=False):

        root = self._convenient_root
        if unmodified:
            root = self._unmodified_root

        return root

    # Similar to to getroot().find(), but includes namespacing support regardless of Python version and
    # automatically adds known namespaces. return_ET will return an ElementTree instance instead of a Label
    def find(self, match, namespaces=None, unmodified=False, return_ET=False):

        # Select the proper XML root to look in
        root = self.getroot(unmodified)

        # Append known namespaces if search contains them
        namespaces = self._append_known_namespaces(match, namespaces)

        # Find the matching element
        try:
            found_element = root.find(match, namespaces=namespaces)

        # Implement namespaces support for Python 2.6
        except TypeError:

            if namespaces is not None:
                match = self._add_namespaces_to_match(match, namespaces)

            found_element = root.find(match)

        # If return_ET is not used, find the other matching element
        if not return_ET and found_element is not None:

            other_element = self._find_other_element(found_element, unmodified)

            if unmodified:
                label = Label(other_element, found_element)
            else:
                label = Label(found_element, other_element)

            found_element = label

        return found_element

    # Similar to to getroot().findall(), but includes namespacing support regardless of Python version and
    # automatically adds known namespaces. return_ET will return an ElementTree instances instead of Labels
    def findall(self, match, namespaces=None, unmodified=False, return_ET=False):

        # Select the proper XML root to look in
        root = self.getroot(unmodified)

        # Append known namespaces if search contains them
        namespaces = self._append_known_namespaces(match, namespaces)

        # Find the matching elements
        try:
            found_elements = root.findall(match, namespaces=namespaces)

        # Implement namespaces support for Python 2.6
        except TypeError:

            if namespaces is not None:
                match = self._add_namespaces_to_match(match, namespaces)

            found_elements = root.findall(match)

        # If return_ET is not used, find the other matching elements
        if not return_ET and found_elements is not None:

            labels = []

            for element in found_elements:
                other_element = self._find_other_element(element, unmodified)

                if unmodified:
                    label = Label(other_element, element)
                else:
                    label = Label(element, other_element)

                labels.append(label)

            found_elements = labels

        return found_elements

    # Returns string representation of label
    def to_string(self, unmodified=False):

        root = self.getroot(unmodified)

        # Obtain string representation, taking care to register_namespaces()
        self._register_namespaces(root, 'register')
        string = ET.tostring(root)
        self._register_namespaces(root, 'unregister')

        return string

    # Returns OrderedDict representation of label
    def to_dict(self, unmodified=False):

        root = self.getroot(unmodified)

        # Achieve equivalent of register_namespaces() via tag_modify
        tag_modify = []
        for uri, prefix in root.namespace_map.items():
            tag_modify.append(('{{{0}}}'.format(uri), '{0}:'.format(prefix)))

        return xml_to_dict(root, tag_modify=tuple(tag_modify))

    # Looks for the same element as given in the other (convenient_root or unmodified_root) root;
    # was_unmodified is True if the input element was taken from unmodified_root
    def _find_other_element(self, element, was_unmodified):

        other_element = None

        if was_unmodified:
            other_root = self.getroot(unmodified=False)
        else:
            other_root = self.getroot(unmodified=True)

        # Loop over all elements to find the matching node_number
        for child_elem in other_root.getiterator():

            if child_elem.node_number == element.node_number:
                other_element = child_elem
                break

        return other_element

    # Registers or unregisters namespaces via ET's register_namespace. The unregister functionality is
    # provided because register_namespace is global, affecting any other ET usage. Register and unregister
    # allow anything in-between them to effectively have a local namespace register.
    @classmethod
    def _register_namespaces(cls, root, action='register'):

        prefixes = ET._namespace_map.values()
        uris = ET._namespace_map.keys()

        for uri, prefix in root.namespace_map.items():

            # Register namespaces
            if action == 'register':

                # Check if namespace prefix or URI already exists, then do not register the namespace if so.
                # This allows local namespaces that have the same prefix as a global prefix but a different URI,
                # which is valid in XML, to stay unique)
                if (prefix in prefixes) or (uri in uris):
                    continue

                # Register the namespace
                ET._namespace_map[uri] = prefix

            # Unregister the namespaces
            else:

                if (prefix in prefixes) and (uri in ET._namespace_map.keys()):
                    del ET._namespace_map[uri]

    # Appends known namespaces (i.e., PDS4_NAMESPACES) to namespaces if match contains prefixes (signified
    # by the colon). In case of duplicates, the 'namespaces' contents overwrite those of PDS4_NAMESPACES
    @classmethod
    def _append_known_namespaces(cls, match, namespaces):

        if (':' in match) and (namespaces is None):
            namespaces = PDS4_NAMESPACES

        elif ':' in match:
            namespaces = dict(chain(PDS4_NAMESPACES.viewitems(), namespaces.viewitems()))

        return namespaces

    # Prepends namespaces to match for Python 2.6; adapted from ET.ElementPath in Python 2.7
    @classmethod
    def _add_namespaces_to_match(cls, match, namespaces):

        xpath_tokenizer_re = re.compile("("
                                        "'[^']*'|\"[^\"]*\"|"
                                        "::|"
                                        "//?|"
                                        "\.\.|"
                                        "\(\)|"
                                        "[/.*:\[\]\(\)@=])|"
                                        "((?:\{[^}]+\})?[^/\[\]\(\)@=\s]+)|"
                                        "\s+")
        modified_match = ''

        for token in xpath_tokenizer_re.findall(match):
            tag = token[1]

            if tag and tag[0] != '{' and ':' in tag:

                try:
                    prefix, uri = tag.split(":", 1)
                    if not namespaces:
                        raise KeyError

                    modified_match += '{0}{{{1}}}{2}'.format(token[0], namespaces[prefix], uri)
                except KeyError:
                    raise SyntaxError('prefix {0} not found in prefix map'.format(prefix))
            else:
                modified_match += '{0}{1}'.format(*token)

        return modified_match
