import os
import sys
import math
import functools
from fractions import Fraction
from Tkinter import Menu, Canvas, Frame, Scrollbar, Label, Entry, Scale, StringVar, BooleanVar, DoubleVar, IntVar
from tkFileDialog import asksaveasfilename

import matplotlib as mpl
import numpy as np

from zscale import zscale
from pds4_viewer import Window, DataViewWindow
from pds4_logging import logger_init

# Initialize TK backend for MPL safely
if mpl.get_backend() != 'TkAgg':

    try:
        mpl.use('TkAgg', force=True)

    # Older MPL versions do not have the force parameter
    except TypeError:
        mpl.use('TkAgg')
        if 'matplotlib.backends' in sys.modules:
            reload(sys.modules['matplotlib.backends'])
else:
    mpl.use('TkAgg')

from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg, NavigationToolbar2TkAgg
from matplotlib.figure import Figure

# Safe import of PowerNorm (available in MPL v1.4+)
try:
    from matplotlib.colors import PowerNorm

except ImportError:
    PowerNorm = None

# Initialize the logger
logger = logger_init()


class ImageViewWindow(DataViewWindow):
    """ Window that displays Array_2D and Array_3D PDS4 data structures as an image.

    This window will display image data. After creating it, you should use the load_array() method to load
    the array that it needs to display.
    """

    def __init__(self, parent):

        # Create basic PDS4 Viewer
        super(ImageViewWindow, self).__init__(parent)

        # Pack the canvas, which contains all UI elements except the menu
        self._canvas.pack(side='left', expand=1, fill='both')

        # Control variable for the _image_display_manage() function
        self._lock_image_display = False

        # The mpl_canvas will be set to an instance of MPL's FigureCanvasTkAgg
        self._mpl_canvas = None

        # Contains sub-widgets of the header widget
        self._header = {'x': None, 'y': None, 'value': None, 'frame': None}

        # Contains most widgets
        self._widgets = {'header': None, 'toolbar': None, 'figure': None, 'image': None,
                         'scrollable_canvas': None, 'colorbar': None, 'colorbar_frame': None,
                         'colorbar_canvas': None}

        # Menu option variables. These are TKinter type wrappers around standard Python variables. The
        # advantage is that you can use trace(func) on them, to call func whenever one of these variables
        # is changed
        update_colorbar = functools.partial(self._update_colorbar, full=True)

        menu_options = [{'name': 'colorbar_show',   'type': BooleanVar(), 'default': True,         'trace': self.toggle_colorbar},
                        {'name': 'colorbar_orient', 'type': StringVar(),  'default': 'horizontal', 'trace': update_colorbar},
                        {'name': 'colormap',        'type': StringVar(),  'default': 'gray',       'trace': self._update_colormap},
                        {'name': 'colormap_invert', 'type': BooleanVar(), 'default': False,        'trace': self._update_colormap},
                        {'name': 'border_show',     'type': BooleanVar(), 'default': False,        'trace': self._update_axis_display},
                        {'name': 'axis_ticks_show', 'type': StringVar(),  'default': 'none',       'trace': self._update_axis_display},
                        {'name': 'scale',           'type': StringVar(),  'default': 'linear',     'trace': self._update_scale},
                        {'name': 'scale_limits',    'type': StringVar(),  'default': 'min max',    'trace': self._update_scale},
                        {'name': 'zoom_level',      'type': DoubleVar(),  'default': 1,            'trace': self._update_zoom},
                        {'name': 'invert_axis',     'type': StringVar(),  'default': 'none',       'trace': self._update_invert_axis},
                        {'name': 'rotation_angle',  'type': DoubleVar(),  'default': 0,            'trace': self._update_rotation},
                        {'name': 'pan',             'type': BooleanVar(), 'default': False,        'trace': self._pan},
                        {'name': 'axis_limits_to_rectangle', 'type': BooleanVar(), 'default': False, 'trace': self._axis_limits_to_rectangle}]

        for option in menu_options:

            var = option['type']
            var.set(option['default'])
            var.trace('w', option['trace'])

            self._menu_options[option['name']] = var

    # Loads the image array structure into this window and displays it for the user
    def load_array(self, array_structure):

        # Orient the array data according to the display dictionary
        try:
            array_data = orient_array(array_structure)

        except TypeError:
            logger.warn('No valid/supported Display Settings found. Displayed image orientation may not be correct.')
            array_data = array_structure.data

        # Set necessary instance variables for this DataViewWindow
        self.data = array_data
        self.meta_data = array_structure.meta_data
        self._settings = {'num_frames': 1, 'display_frame': 0, 'dpi': 80.,
                          'pixel_init_dimensions': (0, 0), 'pixel_margins': (0, 0)}

        if 'Array_3D_' in array_structure.type:
            image_data = array_data[0]
            self._settings['num_frames'] = len(array_data)
        else:
            image_data = array_data

        # Create the header
        self._draw_header()

        # Create the scrollable canvas that will contain the figure
        scrollable_canvas = Canvas(self._display_frame, takefocus=False, bd=0, highlightthickness=0)
        scrollable_canvas.place(relx=0.5, rely=0.5, anchor='center')
        self._widgets['scrollable_canvas'] = scrollable_canvas

        # Create the scrollbars for our scrollable scrollable_canvas
        self._vert_scrollbar = Scrollbar(self._canvas, orient='vertical', command=scrollable_canvas.yview)
        scrollable_canvas.config(yscrollcommand=self._vert_scrollbar.set)
        self._vert_scrollbar.pack(side='right', fill='y')

        self._horz_scrollbar = Scrollbar(self._canvas, orient='horizontal', command=scrollable_canvas.xview)
        scrollable_canvas.config(xscrollcommand=self._horz_scrollbar.set)
        self._horz_scrollbar.pack(side='bottom', fill='x')

        # Create the figure and add image to it
        self._load_image(image_data)

        # Add notify event for window resizing
        scrollable_canvas.bind('<Configure>', self._window_resize)

        # Add notify event for scroll wheel (used to zoom via scroll wheel)
        self._bind_scroll_event(self._mousewheel_scroll)

        # Add notify event for mouse pointer exiting scrollable_canvas (used to display pixel values under pointer)
        self._mpl_canvas.get_tk_widget().bind('<Leave>', self._mouse_pixel_value_adjust)

        # Add notify event for mouse motion (used to display pixel values under pointer)
        self._mpl_canvas.mpl_connect('motion_notify_event', self._mouse_pixel_value_adjust)

        self._add_menu()
        self._data_open = True

    # Sets colormap on currently displayed image
    def set_colormap(self, colormap):
        self._menu_options['colormap'].set(colormap)

    # Controls scaling for the currently displayed image. scale can have values linear|log|squared|square root,
    # and limits can have values min max|zscale. To keep existing values use None.
    def set_scale(self, scale=None, limits=None):

        if scale is not None:
            self._menu_options['scale'].set(scale)

        if limits is not None:
            self._menu_options['scale_limits'].set(limits)

    # Zooms in on the initial image dimensions by zoom_level (e.g. zoom_level of 2 will double the dimensions
    # of the original image)
    def set_zoom(self, zoom_level):
        self._menu_options['zoom_level'].set(zoom_level)

    # Rotates currently displayed image to rotation_angle from initial orientation, in degrees. Note that
    # rotation angle must be evenly divisible by 90.
    def set_rotation(self, rotation_angle):
        self._menu_options['rotation_angle'].set(rotation_angle)

    # Controls axes display for current image. show_axis and invert_axis can have values x|y|xy|none.
    # show_border is a boolean that controls whether a border is displayed around image. To keep existing
    # values use None.
    def set_axes_display(self, show_axis=None, invert_axis=None, show_border=None):

        if show_axis is not None:
            self._menu_options['axis_ticks_show'].set(show_axis)

        if invert_axis is not None:
            self._menu_options['invert_axis'].set(invert_axis)

        if show_border is not None:
            self._menu_options['border_show'].set(show_border)

    # Zooms the image, via resampling, to fit the current window
    def zoom_to_fit(self):

        fig_size = self._settings['pixel_init_dimensions']
        margins = self._settings['pixel_margins']

        # Retrieve width and height of canvas_frame (this is the width and height that we need to zoom to fit)
        self._widget.update_idletasks()
        width = self._display_frame.winfo_width()
        height = self._display_frame.winfo_height()

        # Adjust maximum zoom level to remove colorbar size (since the colorbar takes up some space we need
        # to fit in)
        if self._widgets['colorbar'] is not None:

            orientation = self._menu_options['colorbar_orient'].get()
            self._widget.update_idletasks()

            if orientation == 'vertical':
                width -= self._widgets['colorbar_frame'].winfo_width()

            else:
                height -= self._widgets['colorbar_frame'].winfo_height()

        # Determine the maximum zoom level possible before image will not fit in frame
        extra_space_width = margins[0] * 2
        extra_space_height = margins[1] * 2

        fig_size_div = ((width - extra_space_width) / fig_size[0],
                        (height - extra_space_height) / fig_size[1])

        zoom_level = min(fig_size_div)
        self._zoom(zoom_level=zoom_level)

    # Updates current image to show a new frame. The frame variable can be set to 'next' or 'previous', or
    # to an integer specifying which frame is to be shown
    def select_frame(self, frame=None):

        last_frame = self._settings['display_frame']

        if isinstance(frame, str):

            if frame == 'next':
                new_frame = last_frame + 1

            elif frame == 'previous':
                new_frame = last_frame - 1

            else:
                raise IndexError('Invalid frame action requested: {0}', frame)

            if (new_frame > self._settings['num_frames'] - 1) or (new_frame < 0):
                new_frame = 0

        else:

            if (not isinstance(frame, (int, long))) or (frame > self._settings['num_frames'] - 1) or (frame < 0):
                raise IndexError('Invalid frame number requested: {0}', frame)

            new_frame = frame

        # Display new frame if needed
        if new_frame != last_frame:
            self._settings['display_frame'] = new_frame

            self._header['frame'].set(new_frame)
            self._update_rotation(reload_image=False)
            self._update_scale()
            self._mpl_canvas.draw()

    # Toggles whether colorbar is displayed for current image
    def toggle_colorbar(self, *args):

        if self._menu_options['colorbar_show'].get():
            self._add_colorbar()

        else:
            self._remove_colorbar()

    # Resets the current view of the image, including all axis-limits, pan, zoom, rotations and inversions,
    # to the original view
    def reset_view(self):

        # Hide the image display temporarily to hide any ugly re-drawing
        self._image_display_manage('hide', lock=True)

        self._zoom(zoom_level=1)
        self.set_axes_display(invert_axis='none')
        self.set_rotation(0)

        # Show the image display again, since it was hidden above
        self._image_display_manage('show', lock=False)

    # Adds menu options used for manipulating the data display
    def _add_menu(self):

        # Adds an Export sub-menu to the File menu
        file_menu = self._menu.winfo_children()[0]
        # export_menu = Menu(file_menu, tearoff=0)

        # file_menu.insert_cascade(0, label='Export', menu=export_menu)
        file_menu.insert_command(0, label='Save Image', command=self._save_file_box)

        file_menu.insert_separator(1)

        # Add a Frame menu
        frame_menu = Menu(self._menu, tearoff=0)
        self._menu.add_cascade(label='Frame', menu=frame_menu)

        frame_menu.add_command(label='Data Cube',
                               command=lambda: self._dependent_windows.append(DataCubeWindow(self._viewer, self)))

        frame_menu.add_separator()

        frame_menu.add_command(label='Next Frame', command=lambda: self.select_frame('next'))
        frame_menu.add_command(label='Previous Frame', command=lambda: self.select_frame('previous'))

        # Add a Zoom menu
        zoom_menu = Menu(self._menu, tearoff=0)
        self._menu.add_cascade(label='Zoom', menu=zoom_menu)

        zoom_menu.add_command(label='Reset Original View', command=self.reset_view)

        zoom_menu.add_separator()

        zoom_menu.add_command(label='Zoom to Fit', command=self.zoom_to_fit)

        zoom_menu.add_separator()

        zoom_menu.add_checkbutton(label='Pan', onvalue=True, offvalue=False, variable=self._menu_options['pan'])
        zoom_menu.add_checkbutton(label='Axis-Limits to Rectangle', onvalue=True, offvalue=False,
                                  variable=self._menu_options['axis_limits_to_rectangle'])

        zoom_menu.add_separator()

        zoom_menu.add_command(label='Zoom In', command=lambda: self._zoom('in'))
        zoom_menu.add_command(label='Zoom Out', command=lambda: self._zoom('out'))

        zoom_menu.add_separator()

        zoom_levels = (1./32, 1./16, 1./8, 1./4, 1./2, 1, 2, 4, 8, 16, 32)

        for level in zoom_levels:
            label = 'Zoom {0}'.format(Fraction.from_float(level))
            zoom_menu.add_checkbutton(label=label, onvalue=level, offvalue=level,
                                      variable=self._menu_options['zoom_level'])

        zoom_menu.add_separator()

        invert_options = ('none', 'x', 'y', 'xy')
        for axis in invert_options:

            if axis == 'none':
                label = 'None'
            else:
                label = 'Invert {0}'.format(axis.upper())

            zoom_menu.add_checkbutton(label=label, onvalue=axis, offvalue=axis,
                                      variable=self._menu_options['invert_axis'])

        zoom_menu.add_separator()

        rotation_angles = (0, 90, 180, 270)

        for angle in rotation_angles:
            label = '{0} Degrees'.format(angle)
            zoom_menu.add_checkbutton(label=label, onvalue=angle, offvalue=angle,
                                      variable=self._menu_options['rotation_angle'])

        # Add a Scale menu
        scale_menu = Menu(self._menu, tearoff=0)
        self._menu.add_cascade(label='Scale', menu=scale_menu)

        scale_types = ('linear', 'log', 'squared', 'square root')

        for scale in scale_types:

            # PowerNorm is an MPL feature available in MPL v1.4+ only
            disable_powernorm = {}

            if (PowerNorm is None) and (scale in ('squared', 'square root')):
                disable_powernorm = {'state': 'disabled'}

            scale_menu.add_checkbutton(label=scale.capitalize(), onvalue=scale, offvalue=scale,
                                       variable=self._menu_options['scale'], **disable_powernorm)

        scale_menu.add_separator()

        scale_limits = ('min max', 'zscale')

        for limit in scale_limits:
            scale_menu.add_checkbutton(label=limit.capitalize(), onvalue=limit, offvalue=limit,
                                       variable=self._menu_options['scale_limits'])

        # Add a Color menu
        color_menu = Menu(self._menu, tearoff=0)
        self._menu.add_cascade(label='Color', menu=color_menu)

        colormaps = ('gray', 'Reds', 'Greens', 'Blues', 'hot', 'afmhot', 'gist_heat', 'cool', 'coolwarm',
                     'jet', 'rainbow', 'hsv')

        for color in colormaps:
            color_menu.add_checkbutton(label=color, onvalue=color, offvalue=color,
                                       variable=self._menu_options['colormap'])

        color_menu.add_separator()

        # Add an All Colors sub-menu to Color menu
        all_colors_menu = Menu(self._menu, tearoff=0)
        color_menu.add_cascade(label='All Colors', menu=all_colors_menu)

        all_colormaps = [m for m in mpl.cm.datad if not m.endswith("_r")]

        for color in all_colormaps:
            all_colors_menu.add_checkbutton(label=color, onvalue=color, offvalue=color,
                                            variable=self._menu_options['colormap'])

        color_menu.add_separator()

        color_menu.add_checkbutton(label='Invert Colormap', onvalue=True, offvalue=False,
                                   variable=self._menu_options['colormap_invert'])

        color_menu.add_separator()

        # Add to Colorbar sub-menu to Color menu
        colorbar_menu = Menu(color_menu, tearoff=0)
        color_menu.add_cascade(label='Colorbar', menu=colorbar_menu)

        colorbar_menu.add_checkbutton(label='Show', onvalue=True, offvalue=False,
                                      variable=self._menu_options['colorbar_show'])

        # Add to Colorbar orientation sub-menu to Colorbar menu
        colorbar_orient_menu = Menu(color_menu, tearoff=0)
        colorbar_menu.add_cascade(label='Orientation', menu=colorbar_orient_menu)

        orientations = ('vertical', 'horizontal')

        for orientation in orientations:
            colorbar_orient_menu.add_checkbutton(label=orientation.capitalize(), onvalue=orientation,
                                                 offvalue=orientation, variable=self._menu_options['colorbar_orient'])

        # Add a View menu
        view_menu = Menu(self._menu, tearoff=0)
        self._menu.add_cascade(label='View', menu=view_menu)

        view_menu.add_checkbutton(label='Show Image Border', onvalue=True, offvalue=False,
                                  variable=self._menu_options['border_show'])

        # Add a Ticks sub-menu to View Menu
        ticks_menu = Menu(view_menu, tearoff=0)
        view_menu.add_cascade(label='Ticks', menu=ticks_menu)

        show_ticks = ('x', 'y', 'xy')
        for axis in show_ticks:
            ticks_menu.add_checkbutton(label='Show {0} Ticks'.format(axis.upper()), onvalue=axis,
                                       offvalue='none', variable=self._menu_options['axis_ticks_show'])

        # Add a Colorbar sub-menu to View menu
        view_menu.add_cascade(label='Colorbar', menu=colorbar_menu)

    # Draws the header box, which contains the name of the structure, the current frame, the pixel mouse is
    # over and that pixel's value
    def _draw_header(self):

        header_box = Frame(self._canvas, bg=self._get_bg('gray'), height=82, borderwidth=1, relief='raised')
        header_box.pack(side='top', fill='x')
        header_box.pack_propagate(False)
        self._widgets['header'] = header_box

        value_box_params = {'background': self._get_bg('gray'), 'borderwidth': 1, 'relief': 'groove'}
        line_box_params = {'background': self._get_bg('gray'), 'height': 25}

        entry_params = {'disabledbackground': self._get_bg('gray'), 'state': 'disabled',
                        'borderwidth': 0, 'highlightthickness': 0, 'width': 10, 'disabledforeground': 'black'}

        # Add the header line containing the structure name
        structure_line_box = Frame(header_box, **line_box_params)
        structure_line_box.pack(side='top', anchor='w', pady=(4, 4))

        w = Label(structure_line_box, text='Structure', bg=self._get_bg('gray'))
        w.pack(side='left', padx=(0, 32))

        structure_name_box = Frame(structure_line_box, height=20, width=330, **value_box_params)
        structure_name_box.pack(side='left')
        structure_name_box.pack_propagate(False)

        w = Label(structure_name_box, text=self.meta_data.id, bg=self._get_bg('gray'))
        w.pack(side='left', padx=(10, 10))

        # Add the header line containing the current frame number
        frame_line_box = Frame(header_box, **line_box_params)
        frame_line_box.pack(side='top', anchor='w', pady=(0, 4))

        w = Label(frame_line_box, text='Frame', bg=self._get_bg('gray'))
        w.pack(side='left', padx=(0, 34))

        frame_num_box = Frame(frame_line_box, height=20, width=330, **value_box_params)
        frame_num_box.pack(side='left')
        frame_num_box.pack_propagate(False)

        frame_num = StringVar()
        frame_num.set('0')
        self._header['frame'] = frame_num
        e = Entry(frame_num_box, textvariable=frame_num, **entry_params)
        e.pack(side='left', padx=(10, 10))

        # Add the header line containing X, Y and pixel value
        pixel_line_box = Frame(header_box, **line_box_params)
        pixel_line_box.pack(side='top', anchor='w', pady=(0, 4))

        w = Label(pixel_line_box, text='Pixel', bg=self._get_bg('gray'))
        w.pack(side='left', padx=(0, 25))

        w = Label(pixel_line_box, text='X', bg=self._get_bg('gray'))
        w.pack(side='left', padx=(0, 5))

        x_pixel_box = Frame(pixel_line_box, height=25, **value_box_params)
        x_pixel_box.pack(side='left')

        x_pixel = StringVar()
        self._header['x'] = x_pixel
        e = Entry(x_pixel_box, textvariable=x_pixel, **entry_params)
        e.pack(side='left', padx=(10, 10))

        w = Label(pixel_line_box, text='Y', bg=self._get_bg('gray'))
        w.pack(side='left', padx=(10, 5))

        y_pixel_box = Frame(pixel_line_box, height=25, **value_box_params)
        y_pixel_box.pack(side='left')

        y_pixel = StringVar()
        self._header['y'] = y_pixel
        e = Entry(y_pixel_box, textvariable=y_pixel, **entry_params)
        e.pack(side='left', padx=(10, 10))

        w = Label(pixel_line_box, text='Value', bg=self._get_bg('gray'))
        w.pack(side='left', padx=(10, 5))

        value_box = Frame(pixel_line_box, height=25, **value_box_params)
        value_box.pack(side='left')

        pixel_value = StringVar()
        self._header['value'] = pixel_value
        e = Entry(value_box, textvariable=pixel_value, **entry_params)
        e.pack(side='left', padx=(10, 10))

    # Initializes or updates the MPL Figure and draws image_data on it. load_array() must have been called
    # for the array this image_data came from at least once prior to this method being called.
    def _load_image(self, image_data):

        # Hide the image display temporarily to hide any ugly drawing
        self._image_display_manage('hide', lock=True)

        # Temporary reset of MPL's rcparams, such that they do not interfere with designed GUI look
        mpl_user_defaults = mpl.rcParams.copy()
        mpl.rcdefaults()

        # Determine image display properties
        dpi = self._settings['dpi']
        x_pixels, y_pixels = image_data.shape[1], image_data.shape[0]
        fig_size = x_pixels / dpi, y_pixels / dpi

        # Create the figure
        if not self._data_open:
            self._widgets['figure'] = Figure(figsize=fig_size, dpi=dpi) # , facecolor='none'

        # Update figure dimensions if it already exists
        else:
            self._remove_colorbar()
            self._widgets['figure'].clear()
            self._set_figure_dimensions(fig_size, type='inches')

        # Store initial figure size
        self._settings['pixel_init_dimensions'] = self._get_figure_dimensions(type='pixels')

        # Take care of miscellaneous initialization
        if not self._data_open:

            # Create the FigureCanvasTkAgg(), which is the MPL TK wrapper for displaying an MPL Figure()
            self._mpl_canvas = FigureCanvasTkAgg(self._widgets['figure'], master=self._widgets['scrollable_canvas'])
            self._widgets['scrollable_canvas'].create_window(0, 0, window=self._mpl_canvas.get_tk_widget(), anchor='nw')
            self._mpl_canvas.get_tk_widget().config(takefocus=False, bd=0, highlightthickness=0)

            # Create the toolbar (hidden, used for its pan and axis-limits to rectangle options)
            self._widgets['toolbar'] = NavigationToolbar2TkAgg(self._mpl_canvas, self._display_frame)
            self._widgets['toolbar'].update()
            self._widgets['toolbar'].pack_forget()

            # "Draw" the image for the very first time (it is not actually displayed since its containing
            # canvas_frame is not packed, and moreover the image itself is empty), and adjust for the MPL
            # issue if necessary
            self._mpl_canvas.draw()

            if self._settings['pixel_init_dimensions'] != self._get_figure_dimensions(type='pixels'):
                self._fix_mpl_size()

        # Create an axis for the new image
        figure_size = self._get_figure_dimensions(type='pixels')
        bbox = self._get_bbox_with_margins(figure_size)

        ax = self._widgets['figure'].add_axes(bbox, axisbelow=True, frame_on=False, clip_on=False)

        # Draw initial image
        self._widgets['image'] = ax.imshow(image_data, origin='lower', interpolation='none',
                                           norm=mpl.colors.Normalize(), aspect='equal')

        # Disable X and Y tick labels initially
        ax.get_xaxis().set_visible(False)
        ax.get_yaxis().set_visible(False)

        # Add colorbar to image
        if self._widgets['colorbar'] is None:
            self._add_colorbar()

        # Once image has been added, update the canvas to account for its size
        self._update_scrollable_canvas_dimensions()

        # Update scale, axis inversion, axis ticks, and zoom to match current menu_options
        self._update_scale()
        self._update_invert_axis()
        self._update_axis_display()
        self._update_zoom()

        # Set MPL's defaults back to previously specified by user
        mpl.rcParams.update(mpl_user_defaults)

        # Show the image display again, since it was hidden above
        self._image_display_manage('show', lock=False)

    # Adds a colorbar to the GUI if menu_options are set to show one
    def _add_colorbar(self):

        # Skip adding a colorbar if none is suppose to be shown
        if not self._menu_options['colorbar_show'].get():
            return

        # Create a new MPL Figure() to for the colorbar
        orientation = self._menu_options['colorbar_orient'].get()
        colorbar_frame = Frame(self._display_frame)
        colorbar_figure = Figure(dpi=self._settings['dpi']) # , facecolor='none'

        # Setup a new axis for the external colorbar
        if orientation == 'vertical':
            colorbar_frame.pack(side='right', fill='y')
            cax = colorbar_figure.add_axes([0.1, 0.05, 0.2, 0.9])

        else:
            colorbar_frame.pack(side='bottom', fill='x')
            cax = colorbar_figure.add_axes([0.04, 0.4, 0.92, 0.4])

        # Create a new MPL TK canvas for the Figure
        colorbar_canvas = FigureCanvasTkAgg(colorbar_figure, master=colorbar_frame)
        self._widgets['colorbar'] = colorbar_figure.colorbar(self._widgets['image'], orientation=orientation, cax=cax, format='%4.2e')

        self._widgets['colorbar_canvas'] = colorbar_canvas
        self._widgets['colorbar_frame'] = colorbar_frame

        # Creating a new scroll event binding because calling FigureCanvasTkAgg() rebinds it
        self._bind_scroll_event(self._mousewheel_scroll)

        # Update colorbar dimensions (in this case setting the colorbar's initial size)
        self._update_colorbar_dimensions()

        colorbar_canvas.draw()
        colorbar_canvas.get_tk_widget().pack()
        colorbar_canvas.get_tk_widget().config(takefocus=False, highlightthickness=0)

        # The image canvas dimensions change due to colorbar taking up space and need updating
        self._update_scrollable_canvas_dimensions()

    # Removes the colorbar from the GUI if one is being shown
    def _remove_colorbar(self, inverse=False):

        # Skip removing the colorbar if none exists
        if self._widgets['colorbar'] is None:
            return

        self._widgets['colorbar_frame'].destroy()
        self._widgets['colorbar'] = None
        self._mpl_canvas.draw()

        # The image canvas dimensions change due to colorbar taking up space and need updating
        self._update_scrollable_canvas_dimensions()

    # Activates the Pan image option
    def _pan(self, *args):

        # Disable axis-limits to rectangle if it is enabled
        limits_option = self._menu_options['axis_limits_to_rectangle']
        pan_option = self._menu_options['pan']

        if pan_option.get() and limits_option.get():
            limits_option.set(False)

        # Enable pan
        self._widgets['toolbar'].pan()

    # Activates the Axis-Limits to Rectangle option
    def _axis_limits_to_rectangle(self, *args):

        # Disable pan if it is enabled
        limits_option = self._menu_options['axis_limits_to_rectangle']
        pan_option = self._menu_options['pan']

        if pan_option.get() and limits_option.get():
            pan_option.set(False)

        # Enable axis-limits to rectangle (called zoom by MPL)
        self._widgets['toolbar'].zoom()

    # Update the image scaling law to match the current menu_options value
    def _update_scale(self, *args):

        type = self._menu_options['scale'].get()
        image_array = self._widgets['image'].get_array()

        if type == 'linear':
            norm = mpl.colors.Normalize()

        elif type == 'log':
            norm = mpl.colors.LogNorm()

        elif type == 'squared':
            norm = PowerNorm(gamma=2.)

        elif type == 'square root':
            norm = PowerNorm(gamma=0.5)

        else:
            raise ValueError('Unknown scale type: {0}'.format(type))

        scale_limits = self._menu_options['scale_limits'].get()

        if scale_limits == 'min max':
            norm.autoscale(image_array)

        elif scale_limits == 'zscale':
            zscale_lim = zscale(image_array)

            norm.vmin = zscale_lim[0]
            norm.vmax = zscale_lim[1]

        else:
            raise ValueError('Unknown scale limits: {0}'.format(scale_limits))

        # Hide the image display temporarily because updating the norm below causes colorbar to twitch
        self._image_display_manage('hide')

        # Update the norm. Calling set_norm() will cause issues when switching norms, because colorbar
        # still has the previous norm, and set_norm() will try to update its min/max without updating the norm
        # type. Therefore we set the norm manually, then update the colormap, then call changed() ourselves on
        # the AxisImage to update it
        self._widgets['image'].norm = norm
        self._update_colormap()
        self._widgets['image'].changed()

        # Show the image display again, since it was hidden above
        self._image_display_manage('show')

        self._mpl_canvas.draw()

    # Update the image rotation to match current menu_options value
    def _update_rotation(self, *args, **kwargs):

        rotation_angle = self._menu_options['rotation_angle'].get()

        # Obtain the initial data
        image_data = self.data
        if len(image_data.shape) == 3:
            image_data = image_data[self._settings['display_frame']]

        if rotation_angle % 90 != 0:
            raise ValueError("Rotation angle '{0}'must be evenly divisible by 90".format(rotation_angle))

        # Rotate the data by intervals of 90 degrees
        rotated_image = np.rot90(image_data, int(rotation_angle / 90))

        # Reload the image (this is necessary to rotate the canvas containing the image)
        if ('reload_image' not in kwargs) or (kwargs['reload_image']):
            self._load_image(rotated_image)

        else:
            self._widgets['image'].set_data(rotated_image)

    # Update the axis inversion to match current menu_options value
    def _update_invert_axis(self, *args):

        invert_axis = self._menu_options['invert_axis'].get()
        axis = self._widgets['image'].axes

        # Reset axes to non-inverted states
        if axis.xaxis_inverted():
            axis.invert_xaxis()

        if axis.yaxis_inverted():
            axis.invert_yaxis()

        # Invert X, Y or both axes
        if invert_axis == 'x':
            axis.invert_xaxis()

        elif invert_axis == 'y':
            axis.invert_yaxis()

        elif invert_axis == 'xy':
            axis.invert_xaxis()
            axis.invert_yaxis()

        elif invert_axis != 'none':
            raise ValueError('Unknown axis to invert: {0}'.format(invert_axis))

        self._mpl_canvas.draw()

    # Update the colormap to match the current menu_options value
    def _update_colormap(self, *args):

        cmap = self._menu_options['colormap'].get()

        if self._menu_options['colormap_invert'].get():
            cmap += '_r'

        self._widgets['image'].set_cmap(cmap)
        self._update_colorbar()
        self._mpl_canvas.draw()

    # Update the colorbar (e.g. if it's been changed from horizontal to vertical, the frame being viewed has
    # changed, or the image's colormap, scaling law, dimensions have been changed)
    def _update_colorbar(self, *args, **kwargs):

        if self._menu_options['colorbar_show'].get():

            # Remove and re-draw if requested (e.g. on change of vertical/horizontal)
            if ('full' in kwargs) and (kwargs['full']):

                # # Hide the image display temporarily to hide any ugly re-drawing
                # self._image_display_manage('hide')

                self._remove_colorbar()
                self._add_colorbar()

                # # Pack the canvas frame which was unpacked above
                # self._image_display_manage('show')

            # Update only colors and ticks otherwise
            else:

                self._update_colorbar_dimensions()

                self._widgets['colorbar'].set_norm(self._widgets['image'].norm)
                self._widgets['colorbar'].set_cmap(self._widgets['image'].cmap)
                self._widgets['colorbar'].update_normal(self._widgets['image'])
                self._widgets['colorbar_canvas'].draw()

        self._mpl_canvas.draw()

    def _update_colorbar_dimensions(self):

        # Skip updating colorbar dimensions if no colorbar is being shown
        if not self._menu_options['colorbar_show'].get():
            return

        orientation = self._menu_options['colorbar_orient'].get()

        # Update idle tasks for accurate winfo_*() results
        self._widget.update_idletasks()

        dpi = self._widgets['colorbar_canvas'].figure.dpi

        if orientation == 'vertical':

            canvas_frame_height = self._display_frame.winfo_height()
            colorbar_size = (1.5 * dpi, canvas_frame_height)

        else:

            canvas_frame_width = self._display_frame.winfo_width()
            colorbar_size = (canvas_frame_width, 0.6 * dpi)

        # Update external colorbar dimensions
        self._set_figure_dimensions(colorbar_size, type='pixels', mpl_canvas=self._widgets['colorbar_canvas'])

        # Update ticks to match new dimensions
        self._widgets['colorbar'].set_ticks(self._get_colorbar_ticks())
        self._widgets['colorbar_canvas'].draw()

    # Shows or hides axis ticks, tick labels and border around image for by menu_options
    def _update_axis_display(self, *args):

        axis = self._widgets['image'].axes

        # Determine axis tick label size appropriate for current image dimensions
        image_array = self._widgets['image'].get_array()
        zoom_level = self._menu_options['zoom_level'].get()

        min_dimension = min(image_array.shape) * zoom_level
        tick_size = int(math.log(min_dimension / 11.) * 4 + 2)

        if tick_size < 2:
            tick_size = 2

        axis.tick_params(axis='both', which='major', labelsize=tick_size)

        # Hide currently shown X and Y ticks
        x_was_visible = axis.get_xaxis().get_visible()
        y_was_visible = axis.get_yaxis().get_visible()

        axis.get_xaxis().set_visible(False)
        axis.get_yaxis().set_visible(False)

        # Show X and Y axis tick labels if requested
        axis_ticks_show = self._menu_options['axis_ticks_show'].get()

        if axis_ticks_show == 'x':
            axis.get_xaxis().set_visible(True)

        elif axis_ticks_show == 'y':
            axis.get_yaxis().set_visible(True)

        elif axis_ticks_show == 'xy':
            axis.get_xaxis().set_visible(True)
            axis.get_yaxis().set_visible(True)

        elif axis_ticks_show != 'none':
            raise ValueError('Unknown axis selected to display: {0}'.format(axis_ticks_show))

        x_is_visible = axis.get_xaxis().get_visible()
        y_is_visible = axis.get_yaxis().get_visible()

        # Adjust margins (for addition/removal of tick labels). We adjust both X and Y margins
        # even if only a single side has tick labels because a tick label close to the side of the image
        # will overflow into the other margin if we adjust only a single one
        pixel_margin = 20 * len(str(max(image_array.shape)))
        change_margin = None

        # Add margin only if: (1) neither x or y already were visible (2) one of them is visible now
        if (not x_was_visible and not y_was_visible) and (x_is_visible or y_is_visible):
            change_margin = pixel_margin
            self._settings['pixel_margins'] = (pixel_margin, pixel_margin)

        # Remove margin only if: (1) either x or y were already visible (2) neither of them is visible now
        elif (x_was_visible or y_was_visible) and (not x_is_visible and not y_is_visible):
            change_margin = -1 * pixel_margin
            self._settings['pixel_margins'] = (0, 0)

        # Adjust margin if necessary (either both tick labels were removed, or at least one was added)
        if change_margin is not None:

            # Calculate new figure dimensions for the pixel margin calculated above
            x_pixels, y_pixels = self._get_figure_dimensions(type='pixels')
            fig_size = (x_pixels + 2 * change_margin, y_pixels + 2 * change_margin)
            bbox = self._get_bbox_with_margins(fig_size)

            # Expand/shrink image dimensions while keeping image the same size via adjusting the margin
            self._widgets['image'].axes.set_position(bbox)
            self._set_figure_dimensions(fig_size, type='pixels')

        # Show / hide image border if requested (works only when tick labels are also shown)
        if self._menu_options['border_show'].get() and (x_is_visible or y_is_visible):
            axis.set_frame_on(True)

        else:
            axis.set_frame_on(False)

        self._mpl_canvas.draw()

    # Update menu_options to specified zoom_level, and then call _update_zoom() to zoom to that level. If
    # no zoom level is specified then an action may be specified, and the image is zoomed in or out (depending
    # on the action) by the zoom_factor variable
    def _zoom(self, action='in', zoom_level=None):

        # Determine zoom level if none if specified
        if zoom_level is None:

            # Obtain the current zoom level
            zoom_level = self._menu_options['zoom_level'].get()

            # Determine zoom level when adjusted by zoom factor
            zoom_factor = 0.20

            if action == 'in':
                zoom_level *= (1 + zoom_factor)
            else:
                zoom_level *= (1 - zoom_factor)

        # Save the new zoom level
        self.set_zoom(zoom_level)

    # Zoom image to the level set by menu_options
    def _update_zoom(self, *args):

        # Hide the image temporarily to hide the ugly re-drawing actions going on
        self._image_display_manage('hide')

        fig_size = self._settings['pixel_init_dimensions']
        margins = self._settings['pixel_margins']

        # Determine what level of zoom to make
        zoom_level = self._menu_options['zoom_level'].get()

        # Change figure size by zoom_level, taking care to prevent margins and colorbar area from being
        # increased by zoom_level
        extra_space_width = margins[0] * 2
        extra_space_height = margins[1] * 2

        zoomed_size = (fig_size[0] * zoom_level + extra_space_width,
                       fig_size[1] * zoom_level + extra_space_height)

        # Adjust margins of figure such that the image it contains will correspond to its zoomed image size
        bbox = self._get_bbox_with_margins(zoomed_size)
        self._widgets['image'].axes.set_position(bbox)

        # Resize the figure to match the zoom level
        self._set_figure_dimensions(zoomed_size, type='pixels')

        # Update colorbar size to match
        self._update_colorbar_dimensions()

        # Update tick sizes match new zoom level
        self._update_axis_display()

        self._mpl_canvas.draw()
        self._image_display_manage('show')

    # Sets the X, Y and pixel value in the header based on current mouse pointer location
    def _mouse_pixel_value_adjust(self, event):

        event_outside_bounds = False

        # If the event is an MPL MouseEvent then we may have scrolled mouse on the image
        if isinstance(event, mpl.backend_bases.MouseEvent):

            image_array = self._widgets['image'].get_array()

            event_has_position = (event.xdata is not None) and (event.ydata is not None)

            is_on_image = event.inaxes == self._widgets['image'].axes
            is_in_x_bounds = (event.xdata >= -0.5) and (image_array.shape[1] > (event.xdata + 0.5))
            is_in_y_bounds = (event.ydata >= -0.5) and (image_array.shape[0] > (event.ydata + 0.5))

            if event_has_position and is_on_image and is_in_x_bounds and is_in_y_bounds:

                x_position = event.xdata + 0.5
                y_position = event.ydata + 0.5
                pixel_value = self._widgets['image'].get_array()[int(y_position), int(x_position)]

                self._header['x'].set('{0:.01f}'.format(x_position))
                self._header['y'].set('{0:.01f}'.format(y_position))
                self._header['value'].set('{0:10}'.format(pixel_value))

            else:
                event_outside_bounds = True

        # Event is not an MPL MouseEvent
        else:
            event_outside_bounds = True

        # Remove values from X, Y and pixel value if the mouse pointer is not on the image
        if event_outside_bounds:
            self._header['x'].set('')
            self._header['y'].set('')
            self._header['value'].set('')

    # Returns ticks for colorbar such that spacing between tick marks is equal and the ticks are able to fit
    # in the current window dimensions
    def _get_colorbar_ticks(self):

        array = self._widgets['image'].get_array()

        norm = self._widgets['image'].norm
        norm.autoscale_None(array)

        orientation = self._menu_options['colorbar_orient'].get()

        colorbar_fig = self._widgets['colorbar_canvas'].figure
        colorbar_size = self._get_figure_dimensions(figure=colorbar_fig, type='inches')

        # Determine how many tick marks to have (double valued function)
        if orientation == 'vertical':
            num_ticks = int(1.5 * colorbar_size[1])

            if num_ticks > 8:
                num_ticks = int(math.log(colorbar_size[1], 1.2))

        else:
            num_ticks = int(0.8 * colorbar_size[0])

            if num_ticks > 8:
                num_ticks = int(math.log(colorbar_size[0], 1.35))

        # Make ticks at even spacial intervals
        edge_remove = 0.12
        ticks = []

        if num_ticks == 1:
            ticks.append(norm.inverse(0.5))

        elif num_ticks > 1:

            increment = (1. / (num_ticks - 1)) * (1. - edge_remove)

            for i in range(0, num_ticks):
                tick_norm = (edge_remove / 2.) + (increment * i)
                ticks.append(norm.inverse(tick_norm))

        return ticks

    # Manage whether image seen on screen by used is displayed, or hidden. If lock is True, any further
    # calls to this function will be ignored unless they specify lock is False.
    def _image_display_manage(self, action='show', lock=None):

        if (not self._lock_image_display) or (lock is False):

            # Hide or show display frame, and the scrollbars. Scrollbars are hidden because their sizes
            # will vary widely as the image is redrawn/resized.
            if action == 'show':
                self._vert_scrollbar.pack(side='right', fill='y')
                self._horz_scrollbar.pack(side='bottom', fill='x')
                self._display_frame.pack(expand=1, fill='both')

                # Update scrollable canvas to account for new scrollregion needed
                self._update_scrollable_canvas_dimensions()

            else:
                self._vert_scrollbar.pack_forget()
                self._horz_scrollbar.pack_forget()
                self._display_frame.pack_forget()

        if lock is not None:
            self._lock_image_display = lock

    # Returns the figure dimensions, in units of type (pixels or inches). By default the image Figure
    # dimensions are returned, however the figure keyword can be used to pass in a different figure
    def _get_figure_dimensions(self, figure=None, type='pixels'):

        if figure is None:
            figure = self._widgets['figure']

        fig_size = figure.get_size_inches()

        if type == 'pixels':
            fig_size = (fig_size[0] * figure.dpi, fig_size[1] * figure.dpi)

        return fig_size[0], fig_size[1]

    # Sets the dimensions of the figure inside mpl_canvas. By default this is the image Figure, however the
    # mpl_canvas keyword can be used to pass a different mpl_canvas.
    def _set_figure_dimensions(self, fig_size, type='pixels', mpl_canvas=None):

        if mpl_canvas is None:
            mpl_canvas = self._mpl_canvas

        if type != 'pixels':
            dpi = mpl_canvas.figure.dpi
            fig_size = (fig_size[0] * dpi, fig_size[1] * dpi)

        class ResizeEvent(object):
            def __init__(self, width, height):
                self.width = width
                self.height = height

        mpl_canvas.get_tk_widget().config(width=fig_size[0], height=fig_size[1])
        mpl_canvas.resize(ResizeEvent(fig_size[0], fig_size[1]))

        self._update_scrollable_canvas_dimensions()

    # Once a matplotlib FigureCanvasTkAgg is drawn via FigureCanvasTkAgg().draw(), but only the first time
    # its drawn, the size of FigureCanvasTkAgg gets resized up by exactly 8 pixels, regardless of the real
    # figure size for some reason. The purpose of this method is to resize it back down by 8 pixels
    def _fix_mpl_size(self):

        fig_size = self._get_figure_dimensions(type='pixels')

        # Remove the extra 8 pixels from figure_size
        fig_size = (fig_size[0] - 8, fig_size[1] - 8)

        # Resize the image back down by 8 pixels
        self._set_figure_dimensions(fig_size, type='pixels')

    # Returns an MPL bounding box such that the MPL axes having this bbox, inside a figure having fig_size,
    # will have margins equivalent to those in self._settings[]
    def _get_bbox_with_margins(self, fig_size):

        x_margin = self._settings['pixel_margins'][0] / fig_size[0]
        y_margin = self._settings['pixel_margins'][1] / fig_size[1]

        bbox = [x_margin, y_margin, 1. - 2. * x_margin, 1. - 2. * y_margin]

        return bbox

    # Updates dimensions of scrollable_canvas to match the current figure size or the window dimensions,
    # whichever is smaller. It also updates the scrollregion of the canvas to match the current figure
    # dimensions, less any space taken up by the colorbar.
    def _update_scrollable_canvas_dimensions(self):

        self._widget.update_idletasks()

        # Retrieve width and height of the image
        width = self._get_figure_dimensions()[0]
        height = self._get_figure_dimensions()[1]

        # Adjust width and height to account for the colorbar, which when present takes up space,
        # requiring that a bigger part of the image needs to be scrollable
        if self._widgets['colorbar'] is not None:

            if self._menu_options['colorbar_orient'].get() == 'vertical':
                width += self._widgets['colorbar_frame'].winfo_reqwidth()

            else:
                height += self._widgets['colorbar_frame'].winfo_reqheight()

        self._widgets['scrollable_canvas'].configure(scrollregion=(0, 0, width, height))

        # Adjust width and height of the scrollable_canvas, while ensuring that it is not larger than the
        # window dimensions (or more specifically the display_frame dimensions, because both header and
        # scrollbars take up some space)
        if width > self._display_frame.winfo_width():
            width = self._display_frame.winfo_width()

        if height > self._display_frame.winfo_height():
            height = self._display_frame.winfo_height()

        self._widgets['scrollable_canvas'].configure(width=width, height=height)

    # Called when the window has been resized, this method updates the dimensions of the various things that
    # need resizing together with a window resize
    def _window_resize(self, event):

        self._update_window_dimensions()
        self._update_scrollable_canvas_dimensions()

        if self._widgets['colorbar'] is not None:
            self._update_colorbar_dimensions()

    # Called on mouse wheel scroll action, zooms image in or out
    def _mousewheel_scroll(self, event):

        # Zoom in
        if event.delta > 0:
            self._zoom('in')

        # Zoom out
        else:
            self._zoom('out')

    #
    def _save_file_box(self):

        filetypes = self._mpl_canvas.get_supported_filetypes().copy()
        default_filetype = self._mpl_canvas.get_default_filetype()

        default_filetype_name = filetypes[default_filetype]
        del filetypes[default_filetype]

        sorted_filetypes = list(filetypes.iteritems())
        sorted_filetypes.sort()
        sorted_filetypes.append((default_filetype, default_filetype_name))

        tk_filetypes = [(name, '*.%s' % ext) for (ext, name) in sorted_filetypes]

        initialdir = self._viewer.last_open_dir
        initialfile = self._mpl_canvas.get_default_filename()

        ### Changing extension broken???

        filename = asksaveasfilename(master=self._widget, title='Save the figure', filetypes=tk_filetypes,
                                     initialdir=initialdir, initialfile=initialfile)

        if filename == '' or filename == ():
            return

        self._viewer.last_open_dir = os.path.dirname(filename)

        # try:
        # This method will handle the delegation to the correct type
        self._mpl_canvas.print_figure(filename, dpi=self._settings['dpi'])
        # except Exception as e:
        #     tkinter_messagebox.showerror("Error saving file", str(e))


class DataCubeWindow(Window):
    """ Window used to show the data cube for an ImageViewWindow """

    def __init__(self, viewer, image_structure_window):

        # Set initial necessary variables and do other required initialization procedures
        super(DataCubeWindow, self).__init__(viewer)

        # Set the title
        self._set_title("{0} - Data Cube".format(self._widget.title()))

        # Set the ImageViewWindow for this DataCube as an instance variable
        self._structure_window = image_structure_window

        # Get basic parameters necessary for DataCubeWindow
        current_frame = image_structure_window.settings['display_frame']
        max_frame = image_structure_window.settings['num_frames'] - 1
        structure_name = image_structure_window.meta_data.id

        # Set the current frame being displayed
        self._current_frame = IntVar()
        self._current_frame.set(current_frame)
        self._current_frame.trace('w', self._slider_moved)

        header_box = Frame(self._widget)
        header_box.pack(anchor='nw', expand=1, fill='x')

        w = Label(header_box, text=structure_name)
        w.pack(anchor='center', pady=(5, 0))

        slider_row_box = Frame(self._widget)
        slider_row_box.pack(pady=(20, 10))

        w = Label(slider_row_box, text=0)
        w.grid(row=1, column=0, padx=(10, 5))

        self._slider = Scale(slider_row_box, from_=0, to=max_frame, length=300, width=15, orient='horizontal',
                             showvalue=0, variable=self._current_frame)
        self._slider.grid(row=1, column=1)

        w = Label(slider_row_box, text=max_frame)
        w.grid(row=1, column=2, padx=(5, 10))

        w = Entry(slider_row_box, bd=0, highlightthickness=0, width=5, justify='center',
                  textvariable=self._current_frame)
        w.grid(row=0, column=1)

    # Calls the ImageViewWindow to update which frame is displayed
    def _slider_moved(self, *args):

        new_frame = self._current_frame.get()
        self._structure_window.select_frame(new_frame)


# Expects a PDS4 2D or 3D array structure with a display settings class. Returns the data of the array,
# modified to have the axis order: [frame_num,] vertical, horizontal. Vertical and horizontal axes are taken
# from the display dictionary of the structure. The frame number will be returned only for a 3D array
def orient_array(image_array_structure):

    array_data = image_array_structure.data
    display_settings = image_array_structure.meta_data.display_settings

    # Return unchanged image if no valid/supported Display Settings exists for this structure
    if (display_settings is None) or (not display_settings.valid):
        raise TypeError('No valid/supported Display Settings found.')

    meta_data = image_array_structure.meta_data
    num_axes = len(meta_data.get_axis_arrays())

    if num_axes > 3:
        raise TypeError('Array contains too many axes ({0}) to orient, expecting max of 3'.format(num_axes))

    # Find names of the vertical and horizontal axes
    display_direction = display_settings['Display_Direction']
    horizontal_name = display_direction['horizontal_display_axis']
    vertical_name = display_direction['vertical_display_axis']

    # Find the Axis_Array for the horizontal and vertical axes
    horizontal_axis = meta_data.get_axis_array(horizontal_name)
    vertical_axis = meta_data.get_axis_array(vertical_name)

    vertical_seq = vertical_axis['sequence_number']
    horizontal_seq = horizontal_axis['sequence_number']

    # Roll the horizontal axis to last position
    array_data = np.rollaxis(array_data, horizontal_seq - 1, num_axes)

    # Adjust vertical_seq to account for horizontal axis moving after it
    if vertical_seq > horizontal_seq:
        vertical_seq -= 1

    # Roll the vertical axis to the 2nd to last position
    array_data = np.rollaxis(array_data, vertical_seq - 1, num_axes - 1)

    # Flip image horizontally if necessary
    if 'Right to Left' in display_direction['horizontal_display_direction']:

        if num_axes == 3:
            array_data = array_data[:, :, ::-1]
        else:
            array_data = array_data[:, ::-1]

    # Flip image vertically if necessary
    if 'Top to Bottom' in display_direction['vertical_display_direction']:

        if num_axes == 3:
            array_data = array_data[:, ::-1]
        else:
            array_data = array_data[::-1]

    return array_data


# Opens a new ImageViewWindow from an ArrayStructure
def open_image(viewer, array_structure):

    image_window = ImageViewWindow(viewer)
    image_window.load_array(array_structure)