import sys
import struct

from pds4_data import get_data_class
from pds4_array_objects import ArrayStructure, Meta_ArrayStructure
from pds4_data_types import data_type_convert_array, get_array_struct_type, adjust_array_type, adjust_complex_type
from pds4_logging import logger_init

# Safe import of numpy (not required)
try:
    import numpy as np
except ImportError:
    np = None

# Initialize the logger
logger = logger_init()

#################################


# Modifies 'data' in-place data values to account for scaling_factor and value_offset
def adjust_data_values(data, scaling_factor, value_offset, use_numpy=False):

    if use_numpy:
        if scaling_factor is not None:
            data *= scaling_factor

        if value_offset is not None:
            data += value_offset

    else:
        for j, datum in enumerate(data):

            if scaling_factor is not None:
                data[j] *= scaling_factor

            if value_offset is not None:
                data[j] += value_offset


# Recursively takes values from data and puts them in a (potentially nested) list having array_dimensions
def _data_to_list(data, array_dimensions, element_offset=0):

    data_list = []

    # Loop over each value in the first array dimension
    for i in range(0, array_dimensions[0]):

        # If array dimension has more than two values (i.e., the current array is at least a 3D array),
        # then recursively call _byte_data_to_list() on each (2D array) element of this 3D array, thereby
        # reducing the number of array dimensions with each recursive call
        if len(array_dimensions) > 2:

            # To actually take the correct values from data (which is done in the below else statement),
            # we need to know which element of data we should taking elements from. The below variable counts
            # how many array elements were already taken
            final_element_offset = element_offset + i * reduce(lambda x, y: x*y, array_dimensions[1:])

            # Reduce via recursion number of array dimensions from len(array_dimensions) to len(array_dimensions) - 1
            reduced_array = _data_to_list(data, array_dimensions[1:], element_offset=final_element_offset)
            data_list.append(reduced_array)

        # Once we reach a 2D array then we can read take each row in that array (this for loop has
        # the effect of adding each row to the data variable, such that data has format data[row][column])
        else:

            # Calculate the first element to be taken for this row
            num_elements = array_dimensions[1]
            final_element_offset = element_offset + i * num_elements

            # Extract from data only the datums for this specific row
            extracted_data_single = data[final_element_offset:final_element_offset+num_elements]

            # Append this row into data_list
            data_list.append(extracted_data_single)

    return data_list


# Modifies 'data' in-place, applying to it the given bit mask
def _apply_bitmask(data, bit_mask_string, data_type):

    # Reverse bit mask string if there is a mismatch between system byte-order and the byte-order of the data
    # as written in the file (this is required because it is the unpacked data that is bitwise operated on)
    struct_type = get_array_struct_type(data_type)[0]
    msb_mismatch = struct_type[0] == '>' and sys.byteorder == 'little'
    lsb_mismatch = struct_type[0] == '<' and sys.byteorder == 'big'

    if msb_mismatch or lsb_mismatch:
        bit_mask_string = bit_mask_string[::-1]

    # Skip needlessly applying bit_mask if it's all 1's
    if '0' not in bit_mask_string:
        return

    # Convert bit mask to binary
    bit_mask = int(bit_mask_string, 2)

    # Apply bit mask to each datum
    for i, datum in enumerate(data):
        data[i] = datum & bit_mask


# Extracts and properly formats the data for a single array, modifies array_structure to contain all extracted data
def _extract_array_data(file_byte_data, array_structure, use_numpy=False, ignore_offset=False):

    # Obtain basic meta data
    meta_data = array_structure.meta_data

    offset = meta_data['offset']
    element_array = meta_data['Element_Array']
    data_type = element_array['data_type']

    # Obtain offset for the data
    if not ignore_offset:
        file_byte_data = file_byte_data[offset:]

    # Sort Axis Arrays by sequence number
    axis_arrays = meta_data.get_axis_arrays()
    sorted_axis_arrays = sorted(axis_arrays, key=lambda x: x['sequence_number'])

    # Get the dimensions that extracted_data will have
    array_dimensions = [axis_array['elements'] for axis_array in sorted_axis_arrays]

    # Convert byte data to its data_type
    num_elements = reduce(lambda x, y: x*y, array_dimensions)
    datum_length = struct.calcsize(get_array_struct_type(data_type)[1])
    data_length = datum_length * num_elements
    converted_data = data_type_convert_array(data_type, file_byte_data[0:data_length], use_numpy)

    # Apply the bit mask to converted_data if necessary
    if ('Object_Statistics' in meta_data) and ('bit_mask' in meta_data['Object_Statistics']):
        bit_mask_string = str(meta_data['Object_Statistics']['bit_mask']).zfill(datum_length*8)
        _apply_bitmask(converted_data, bit_mask_string, data_type)

    # Make scaling_factor and value_offset adjustments
    scaling_factor = element_array.get('scaling_factor')
    value_offset = element_array.get('value_offset')

    if (scaling_factor is not None and scaling_factor != 1) or (value_offset is not None and value_offset != 0):

        # If 'scaling_factor' or 'value_offset' exist then we need to adjust array's type if the types do not match
        converted_data = adjust_array_type(converted_data, scaling_factor, value_offset, use_numpy)

        # Adjust data values to account for 'scaling_factor' and 'value_offset'
        adjust_data_values(converted_data, scaling_factor, value_offset, use_numpy)

    # Adjust data to use complex data types if it had them
    if 'complex' in data_type:
        converted_data = adjust_complex_type(converted_data, use_numpy)

    # Extract data into properly formatted list
    extracted_data = converted_data
    if len(array_dimensions) > 1:
        extracted_data = _data_to_list(converted_data, array_dimensions)

    # Create the Data object
    extracted_data = get_data_class(extracted_data, meta_data, use_numpy)

    # Modify array_structure to contain the extracted and formatted data
    array_structure.data = extracted_data


# Create the array structure, containing label, data and meta data
def get_array(full_label, array_label, file_byte_data, data_filename, use_numpy=False, structure_num=None):

    array_xml = array_label.getroot()

    # Skip over data structure if its not actually an Array
    if 'Array' not in array_xml.tag:
        raise TypeError('Attempted to get_array() on a non-array: ' + array_xml.tag)

    # Ignore offset when extracting data in the case of structure_num being specified (because the data has
    # been read starting from that offset already)
    ignore_offset = False
    if structure_num is not None:
        ignore_offset = True

    # Create the meta data structure for this array
    meta_array_structure = Meta_ArrayStructure()
    meta_array_structure.load(array_xml, full_label)

    # Create the data structure for this array
    array_structure = ArrayStructure(data_filename, array_label, meta_array_structure)

    logger.info('Now processing a {0} structure: {1}'.format(array_xml.tag, array_structure.id))

    # Read the array from file into the array_structure variable
    _extract_array_data(file_byte_data, array_structure, use_numpy, ignore_offset)

    return array_structure

